/* Modified for IBM PC/XT/AT, 9/12/89..9/01/90 from file ami.c:
             Peter C. Lind, M.Sc.,   lind@maccs.ca
             McMaster University,
             1280 Main St. W.,
             Hamilton, ON
             Canada   L8S 4K1

   This file contains only functions which are specific to the IBM PC/XT/AT
   under Turbo C 2.00+.  It connects to the other source files via a small
   number of functions. main() is here.  It calls init() and onestep().
   The other source files can call rnd(), draw() and panic() from this file;
   these are at the end of the file. */

#include "const.h"
#include "var.h"

#include <conio.h>
#include <dos.h>
#include <graphics.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#define TRUE !0
#define FALSE 0

#define STDOUT_HANDLE 1                  /* DOS file handle for stdout */

extern unsigned long RangeSeed;          /* TC random number generator */
extern unsigned char t[2][MAXX][MAXY];   /* Defined in tec1.c */
extern short step;                       /* Also defined in tec1.c */

unsigned long         class;
unsigned short        code;

/* EGA: 16 colors used: Black, 2 shades of blue (dark, light), 2 of green,
        2 of red, 2 of purple (magenta), 2 white (light gray, white), and
	2 yellow (brown, yellow). */
#define NUMCOLORS 13
#define MAXCOLOR 12
unsigned short colors [NUMCOLORS] = {
                                     BLACK,
                                     BLUE,LIGHTBLUE,GREEN,LIGHTGREEN,
                                     YELLOW,MAGENTA,LIGHTMAGENTA,
                                     RED,LIGHTRED,BROWN,
                                     LIGHTGRAY,WHITE
                                    };

unsigned short CellXsize,CellYsize;   /* Size of a tec pixel */
int MaxX,MaxY,CentX,CentY;            /* Screen parameters */
int LastVideoMode;                    /* Video mode at program start up */
int gdriver;                          /* Graphics driver used for plotting */
int BrightColor,DimColor,BorderColor,BackgroundColor,BarColor;
int Bar3Dcolor,PlotBackColor;
int wx1,wy1,wx2,wy2,winMaxX,winMaxY,origY2;  /* Map window coords */

int LegCellWid,LegCellHgt,LegHalfWid; /* Legend parameters */
int legX1,legY1,legX2,legY2,legCY;

short InColor;                        /* TRUE if mode is EGA or VGA */
short Mode3D;                         /* TRUE when plotting in 3d mode */
short Slow3D;                         /* TRUE for slow 3d plotting */
short AllStop;                        /* TRUE if user wants to quit */
short WantPrint;                      /* TRUE if user wants to print */
short WantPause;                      /* TRUE if user wants to pause */
short WantChangePlot;                 /* TRUE if user wants to change
                                          plotting style */
short NeedLegend;                     /* TRUE if legend should be drawn */
short NoLegend;                       /* TRUE if legend should be erased */

int msgY;                             /* Y coord of message line top */

#define xaddon 5                      /* Width of 3-d bars */
int ix;                               /* Initial X coord for 3-d */
int iy;                               /* Initial Y coord for 3-d */
int yaddon;                           /* Y increment for 3-d plotting */

#define ESC 27                        /* ASCII code for ESC */

#define NUM_CREDITS 5
static char title[]   = "TEC   Continental Drift Simulator";
static char credit[NUM_CREDITS][42] =
             {
              "Version  3",
              "originally written by David Allen",
              "adapted for IBM PC/XT/AT by Peter C. Lind",
              "Copyright (c) 1989,1990",
              "press any key to begin"
             };
int CredFont,TitleSize,CredSize;

void TakeMax (a,b)
 /* `a' gets maximum of `a' and `b'. */
 int *a,b;
  {
   if (b > *a)
      *a = b;
  }

void panic (s)
 char *s;
 /* Used when some fatal inconsistency is found in the database;
    its function is to immediately free all graphics memory and exit. */
  {
   closegraph();
   textmode(LastVideoMode);
   printf("TEC PANIC: %s\n",s);
   exit(0);
  }

void message (s)
 char *s;
 /* Displays message `s' at the bottom of the screen. */
  {
   struct viewporttype ov;

   getviewsettings(&ov);                       /* Save current viewport */
   setviewport(0,0,MaxX,MaxY,TRUE);            /* Select entire screen */
   setfillstyle(SOLID_FILL,BarColor);
   bar(0,msgY,MaxX,MaxY);                      /* Blank bottom line */
   settextjustify(CENTER_TEXT,TOP_TEXT);
   setcolor(DimColor);
   outtextxy(CentX,msgY+1,s);                  /* Output message */
   setviewport(ov.left,ov.top,ov.right,        /* Select old viewport */
               ov.bottom,ov.clip);
  }

void WaitKeyPress ()
 /* Wait for user to press a key. */
  {
   while (!kbhit())
         ;
   if (!(getch()))              /* Consume any extended code */
      getch();
  }

void Intro ()
 /* Displays the introductory credits panel. */
  {
   int th,tw,x1,y1,x2,y2,y,ht,hc,cx,i,TwoEms;

   /* Calculate dimensions of panel required */
   settextstyle(CredFont,HORIZ_DIR,TitleSize);    /* Measure title line */
   ht = textheight(title); th = (ht * 3);
   tw = textwidth(title) + (textwidth("M") << 1);
   settextstyle(CredFont,HORIZ_DIR,CredSize);     /* Measure credit lines */
   TwoEms = textwidth("M") << 1;
   hc = 0;
   for (i = 0; i < NUM_CREDITS; i++)              /* Find max. height, width */
       {
	TakeMax(&hc,textheight(credit[i]));
        TakeMax(&tw,textwidth(credit[i])+TwoEms);
       }
   hc <<= 1;
   th += (hc * NUM_CREDITS);
   x1 = (MaxX - tw) >> 1; x2 = x1 + tw - 1;
   y1 = (MaxY - th) >> 1; y2 = y1 + th - 1;
   cx = (tw >> 1) - 1;

   /* Draw panel and display title and credits */
   setfillstyle(SOLID_FILL,BarColor); setcolor(BarColor);
   bar3d(x1,y1,x2,y2,4,TRUE);
   setviewport(x1+1,y1+1,x2-1,y2-1,TRUE);
   y = ht;
   settextstyle(CredFont,HORIZ_DIR,TitleSize);
   settextjustify(CENTER_TEXT,TOP_TEXT);
   setcolor(BrightColor);
   outtextxy(cx,y,title);
   y += (ht << 1);
   settextstyle(CredFont,HORIZ_DIR,CredSize);
   for (i = 0; i < NUM_CREDITS; i++, y += hc)
       outtextxy(cx,y,credit[i]);
   WaitKeyPress();
   setviewport(0,0,MaxX,MaxY,TRUE);
   cleardevice();
  }

void CGAdefaults ()
 /* Common routine to set CGA/MCGA defaults. */
  {
   BarColor = 1; BrightColor = 0; DimColor = 0;
   BorderColor = 1; BackgroundColor = 0; Bar3Dcolor = 1; PlotBackColor = 0;
   InColor = FALSE; Mode3D = TRUE; Slow3D = FALSE;
  }

int initCGA ()
 /* Initialize CGA mode. */
  {
   CGAdefaults();
   CredFont = SMALL_FONT; TitleSize = 6; CredSize = 4;
   return(CGAHI);
  }

int initMCGA ()
 /* Initialize MCGA mode. */
  {
   CGAdefaults();
   CredFont = SMALL_FONT; TitleSize = 6; CredSize = 4;
   return(MCGAHI);
  }

void EGAdefaults ()
 /* Common routine to set EGA/VGA defaults. */
  {
   BarColor = DARKGRAY; BrightColor = WHITE; DimColor = LIGHTGRAY;
   PlotBackColor = BLACK;
   BorderColor = DARKGRAY; BackgroundColor = BLACK; Bar3Dcolor = LIGHTGRAY;
   InColor = TRUE; Mode3D = FALSE;
   LegCellWid = 20; LegHalfWid = 10;
   NeedLegend = TRUE;
  }

int initEGA ()
 /* Initialize EGA */
  {
   EGAdefaults();
   CellXsize = 7; CellYsize = 3;
   CredFont = TRIPLEX_FONT; TitleSize = 3; CredSize = 2;
   LegCellHgt = 16;
   return(EGAHI);
  }

int initVGA ()
 /* Initialize VGA */
  {
   EGAdefaults();
   CellXsize = 7; CellYsize = 5;
   CredFont = TRIPLEX_FONT; TitleSize = 4; CredSize = 3;
   LegCellHgt = 16;
   return(VGAHI);
  }

void grafinit ()
 /* Detect the video hardware and configure accordingly; display credits */
  {
   int gm,tw,th,rc;

   LastVideoMode = LASTMODE;
   detectgraph(&gdriver,&gm);
   switch (gdriver)
	  {
           case CGA: gm = initCGA();   /* CGA high-res mode: 640 x 200, 2 colors */
                     break;
           case MCGA: gm = initMCGA(); /* MCGA high-res mode: 640 x 480, 2 colors */
                      break;
	   case EGA: gm = initEGA();   /* EGA high-res mode: 640 x 350, 16 colors */
                     break;
	   case VGA: gm = initVGA();   /* VGA high-res mode: 640 x 480, 16 colors */
                     break;
           default: printf("TEC 3:\n");
                    printf("Unable to operate with detected display type\n");
                    printf("Require CGA, MCGA, EGA or VGA to run\n");
                    exit(0);
	  }
   initgraph(&gdriver,&gm,"");
   if ((rc = graphresult()))
      {
       printf("TEC 3:\n");
       printf("BGI Error #%d: %s\n",rc,grapherrormsg(rc));
       textmode(LastVideoMode);
       exit(0);
      }
   MaxX = getmaxx(); MaxY = getmaxy();
   CentX = MaxX >> 1; CentY = MaxY >> 1;
   Intro();                                             /* Intro panel */
   settextstyle(DEFAULT_FONT,HORIZ_DIR,1);              /* Choose font */
   tw = textwidth("M"); th = textheight("M");           /* Font's dimensions */
   msgY = MaxY - th - 2;

   setfillstyle(SOLID_FILL,BarColor); bar(0,0,MaxX,th+1); /* Top bar */
   bar(0,msgY,MaxX,MaxY);                                 /* Bottom bar */
   setcolor(BrightColor);
   settextjustify(LEFT_TEXT,TOP_TEXT);
   moveto(tw,1); outtext("TEC ");
   setcolor(DimColor); outtext("Continental Drift Simulator");
   settextjustify(RIGHT_TEXT,TOP_TEXT);
   outtextxy(MaxX-tw,1,"David Allen, Peter C. Lind");

   setcolor(BorderColor); rectangle(0,th+1,MaxX,msgY);    /* Map border */
   wx1 = 2; wy1 = th + 3; wx2 = MaxX - 2; wy2 = msgY - 2;
   winMaxX = wx2 - wx1; winMaxY = wy2 - wy1; origY2 = wy2;

   legX2 = winMaxX; legX1 = legX2 - (LegCellWid * NUMCOLORS); /* Legend coords */
   legY2 = winMaxY - 1; legY1 = legY2 - LegCellHgt;
   legCY = (legY2 + legY1) >> 1;

   setviewport(wx1,wy1,wx2,wy2,TRUE);

   ix = (wx2 - wx1) - (XSIZE * xaddon);
   yaddon = ((wy2 - wy1 + 1) / YSIZE) + 1;
   iy = (wy2 - wy1 + 1) - (YSIZE * yaddon);
  }

void grafexit ()
 /* Just close all the things that were opened, then exit(). */
  {
   closegraph();
   textmode(LastVideoMode);
   printf("TEC ended. Bye\n");
   exit(0);
  }

unsigned GetIOCTL (fh)
 /* Returns the IOCTL word for file handle `fh'. */
 int fh;
  {
   struct REGPACK regs;

   regs.r_ax = 0x4400;            /* DOS Int. sub-function: "Get IOCTL info" */
   regs.r_bx = fh;                /* File handle in BX */
   intr(33,&regs);                /* Call DOS */
   return(regs.r_dx);             /* IOCTL info in DX */
  }

void PrintMenu ()
 /* Displays the print menu. */
  {
   message("Print to stdout:  G)eneric  T)ext  P)ostscript  ESC=No print");
  }

void TryToPrint ()
 /* Attempts to initiate printing, but first warns the user that stdout
    must have been redirected on the DOS command line. Also allows the
    user to choose the print format:

                   G    Generic
                   T    Text
                   P    Postscript
                  ESC   No print
 */
  {
   /* Disallow printing if stdout is still directed to the console screen,
      (ie. IOCTL for stdout indicates device with standard console output). */
   if ((GetIOCTL(STDOUT_HANDLE) & 0x0082) == 0x0082)
      {
       message("*** stdout not redirected.  [ Press any key ] ***");
       do {                        /* Wait for keypress */
          } while (!kbhit());
       if (!getch())               /* Consume character */
          getch();                 /*   and extended code, if any */
       return;
      }

   PrintMenu();
   do {
       char c;

       if (!(c = getch()))
          getch();
         else
          switch (c)
                 {
                  case ESC: return;
                            break;
                  case 'g':
                  case 'G': message("Sending Generic Mode to stdout...");
                            tecst(step % 2,DRAWMODE_GENERIC);
                            PrintMenu();
                            break;
                  case 't':
                  case 'T': message("Sending Text Mode to stdout...");
                            tecst(step % 2,DRAWMODE_TEXT);
                            PrintMenu();
                            break;
                  case 'p':
                  case 'P': message("Sending Postscript Mode to stdout...");
                            tecst(step % 2,DRAWMODE_GRAY);
                            PrintMenu();
                            break;
                  default: break;
                 }
      } while (TRUE);
  }

void Set3Dstyle ()
 /* Allows user to choose fast or slow 3-D plotting. */
  {
   short OK;

   Mode3D = TRUE;
   message("3-D Style: S)low  F)ast");
   do {
       OK = TRUE;
       switch (getch())
              {
               case 's':
               case 'S': Slow3D = TRUE;
                         break;
               case 'f':
               case 'F': Slow3D = FALSE;
                         break;
               default: OK = FALSE;
                        break;
              }
      } while (!OK);
  }

void ChangePlotStyle ()
 /* Allows user to switch between standard plotting and 3-D plotting. Should
    only be called if program is running under EGA/VGA. */
  {
   short OK;

   message("Switch plotting style:  S)tandard  3)D");
   do {
       OK = TRUE;
       switch (getch())
              {
               case '\0': getch();
                          break;
               case '3': NoLegend = !Mode3D;
                         Set3Dstyle();
                         break;
               case 's':
               case 'S': NeedLegend = Mode3D;
                         Mode3D = FALSE;
                         break;
               default: OK = FALSE;
                        break;
              }
      } while (!OK);
  }

void checkmouse ()
 /* Standard event handler. Currently on the IBM PC/XT/AT, can only poll the
    keyboard for user actions (no mouse support [yet!]):

           <ESC>      Quit program right away
          <SPACE>     Pause
          g or G      Change plotting style
          p or P      Print out the current values of t[src]
                      using tecst() in tec1.c. */
  {
   short i,j,k;
   char c;

   if (!kbhit())                  /* Key pressed? */
      return;                     /*   No: Exit */
   if (!(c = getch()))            /* Read the key. Is it '\0'? */
      {
       getch(); return;           /*   Yes: Consume extended code and exit */
      }
   switch (c)                     /*   No: Look at key pressed */
          {
           case ESC: grafexit();
                     break;
           case ' ': WantPause = TRUE;
                     break;
           case 'g':
           case 'G': if (InColor)
                        ChangePlotStyle();
                       else
                        Set3Dstyle();
                     break;
           case 'p':
           case 'P': TryToPrint();
                     break;
           default: break;
          }
  }

int rnd (top)
 int top;
 /* Returns a random number in the range 0..`top'-1. */
  {
   return(random(top));
  }

void ResetPlotSettings ()
 /* Resets the 3-D plot settings (used after message() calls). */
  {
   if (Mode3D)
      {
       setfillstyle(SOLID_FILL,PlotBackColor);
       setcolor(Bar3Dcolor);
      }
  }

void PlotMessage ()
 /* Prints the plot message */
  {
   message("Plotting...  G)raphing style  P)rint  CR=Skip  SP=Pause  ESC=Quit");
  }

int PlotKeyCheck ()
 /* Called when a key is pressed while plotting. If the key was ESC or
    SPACE, then TRUE is returned. Otherwise, FALSE is returned. If the
    key pressed was ESC, 'p' or 'P', the key is not consumed. */
  {
   char c;

   switch (c = getch())
          {
           case '\0': getch();            /* Discard extended code */
                      break;
           case ESC: AllStop = TRUE;      /* Set quit flag... */
           case '\r': return(TRUE);       /* Exit plot routine */
                      break;
           case ' ': WantPause = TRUE;
                     break;
           case 'g':
           case 'G': if (InColor)         /* Defer in EGA/VGA */
                        WantChangePlot = TRUE;
                       else
                        {                 /* Change on the fly in CGA/MCGA */
                         Set3Dstyle();
                         PlotMessage();
                         ResetPlotSettings();
                        }
                     break;
           case 'p':
           case 'P': WantPrint = TRUE;    /* Set print request flag */
                     break;
           default: break;                /* Ignore all else */
          }
   return(FALSE);
  }

void draw3d (src)
 /* Draws a 3-D picture of the t[src] in the center of the current viewport
    using Turbo C's bar3d() function. When `Slow3D' is TRUE, each "pixel" of
    t[src] is printed as a standing 3-D block; when `Slow3D' is FALSE,
    adjacent "pixels" in the same row with equal heights are merged and
    plotted as a wider standing 3-D block, thereby speeding up the plotting
    process (with a loss of attractiveness). */
 short src;
  {
   int boty,x1,x2,x3,x,y,cx,cy,h,nh,c;

   ResetPlotSettings();

   x = ix; y = iy;
   top: for (cx = 0; cx < XSIZE; cx++, y += yaddon)
            {
             x1 = x; x2 = x1 + xaddon; x3 = x + (xaddon * XSIZE);
             h = -1; c = 0;
             for (cy = 0; cy < YSIZE; cy++)
                 {
                  nh = t[src][cx][cy] >> 3;
                  if ((nh != h) || Slow3D)
                     {
                      if ((c || Slow3D) && (h >= 0))
                         bar3d(x1,y,x2,y-h,3,TRUE);
                      h = nh; x1 = x2; c = 1;
                      if (Slow3D || 1)
                         x2 += xaddon;
                     }
                    else
                     {
                      x2 += xaddon; c++;
                     }
                 }
             if (c)
                bar3d(x1,y,x3,y-nh,3,TRUE);
             x -= (xaddon >> 1);

             if (kbhit())
                if (PlotKeyCheck())
                   return;
            }
  }

void DisplayLegend ()
 /* Draws the colors legend for EGA/VGA modes. */
  {
   int i,x;
   char num[3];

   settextstyle(DEFAULT_FONT,HORIZ_DIR,1);
   settextjustify(CENTER_TEXT,CENTER_TEXT);

   /* Special case for black color */
   x = legX1;
   setcolor(DARKGRAY);
   setfillstyle(SOLID_FILL,colors[0]);
   bar3d(x,legY1,x+LegCellWid,legY2,0,FALSE);
   outtextxy(x+LegHalfWid,legCY,"0");
   x += LegCellWid;

   /* Draw other colors in a loop, with black numbers */
   setcolor(BLACK);
   for (i = 1; i < NUMCOLORS; i++, x += LegCellWid)
       {
        setfillstyle(SOLID_FILL,colors[i]);
        bar3d(x,legY1,x+LegCellWid,legY2,0,FALSE);
        sprintf(num,"%d",i);
        outtextxy(x+LegHalfWid,legCY,num);
       };
   wy2 = origY2 - LegCellHgt - 2;       /* Resize viewport to protect legend */
   setviewport(wx1,wy1,wx2,wy2,TRUE);
  }

void RemoveLegend ()
 /* Effectively removes the colors legend by resetting the map viewport to
    full size. */
  {
   wy2 = origY2; setviewport(wx1,wy1,wx2,wy2,TRUE);
  }

void draw (src)
 short src;
 /* When `Mode3D' is TRUE, this function calls draw3d() and exits. Otherwise,
    this function takes the array m[src] and draws it on the screen, in a
    hopefully efficient way.  The function tries to make long horizontal
    patches that are all the same color.  Then they can be rendered by a
    graphics function that draws arbitrary rectangles quickly. */
  {
   register short i,j,k,x;

   checkmouse();
   PlotMessage();

   /* Erase the viewport first */
   if (NoLegend)
      {
       RemoveLegend(); NoLegend = FALSE;
      }
   setcolor(BackgroundColor); clearviewport();

   if (Mode3D)                  /* If in 3-D mode, then call draw3d() */
      {
       draw3d(src); return;
      }

   /* Draw color legend whenever it needs to be displayed. */
   if (NeedLegend)
      {
       DisplayLegend(); NeedLegend = FALSE;
      }

   /* For each scan line, start at the left edge */
   for (j = 0, i = 0; j < YSIZE; j++, i = 0)
       {
        int x1,y1,x2,y2;

	/* If the current square is not ocean, set x to its color */
	top: if ((x = t[src][i][j] >> 2) > 1)
                {
                 /* Go as far along to the right as you can in this color */
                 k = i + 1;
                 while (((t[src][k][j] >> 2) == x) && (k < XSIZE-1))
                       k++;

                 /* Draw a short, wide rectangle */
                 if (x > 27)
                    x = 27;
                 x1 = i * CellXsize; y1 = j * CellYsize;
                 x2 = k * CellXsize; y2 = y1 + CellYsize - 1;
                 setfillstyle(SOLID_FILL,colors[x]);
                 bar(x1,y1,x2,y2);
                 i = k-1;

                 if (kbhit())
                    if (PlotKeyCheck())
                       return;
                }

        /* If not at end of scanline, do more; else start next scanline */
        if (i < XSIZE-1)
           {
            i++;
            goto top;
           }
       }
  }

void main (argc,argv)
 /* Initializes everything and enters a (finite) loop that repeatedly calls
    onestep() and checkmouse() until the maximum step is reached or the user
    quits. */
 int argc;
 char **argv;
  {
   char menu[80];

   randomize();     /* Randomize random number generator using DOS time */

   /* Initialize everything */
   AllStop = WantPrint = WantPause = WantChangePlot = FALSE; /* Clear deferrals */
   NoLegend = FALSE;
   grafinit();
   message("Setting up...");
   init(*++argv);                 /* Perform major initialization */
   checkmouse();                  /* Give user a chance to do something */

   /* Call onestep() once per step; then check for user key actions */
   for (step = 0; step < MAXSTEP; step++)
       {
        /* Check for deferred commands */
        if (AllStop)
           grafexit();
        if (WantPrint)
           TryToPrint();
        if (WantChangePlot)
           ChangePlotStyle();
        if (WantPause)
           {
            message("[ Pausing -- Press any Key ]");
            WaitKeyPress();
           }
        WantPrint = WantPause = WantChangePlot = FALSE;  /* Clear deferrals */

        /* Do some work */
        sprintf(menu,"Thinking [%d]... G)raphing style  P)rint  SP=Pause  ESC=Quit",
                     step);
	message(menu);
        onestep();
        checkmouse();
       }

   /* Reached end of simulation without error or user quit signal. */
   message("Execution terminated -- Press any key");

   WaitKeyPress();              /* Loop forever until user pressed a key */

   closegraph();                /* Shut down */
   textmode(LastVideoMode);
  }
